/*
 * Curses driver by Michael Hope <michaelh@juju.net.nz>, 21 Jan 2001
 *
 * With some addtional hacking by Nils M Holm:
 * - Added support for SHIFT and CONTROL.
 * - Added color support.
 * - Added support for function keys (arrows, PGUP/DN, F-keys, etc).
 */

#include <stdlib.h>
#include <string.h>

#include "global.h"

#include "vgahard.h"
#include "hardware.h"
#include "bios.h"

#include <curses.h>

static WINDOW *scr;

void curses_end_video(void)
{
	endwin();
}

void curses_start_video(void)
{
	int	colors[] = { COLOR_BLACK, COLOR_BLUE, COLOR_GREEN,
				COLOR_CYAN, COLOR_RED, COLOR_MAGENTA,
				COLOR_YELLOW, COLOR_WHITE };
	int	f, b;

	scr = initscr();
	start_color();
	for (b=0; b<8; b++) {
		for (f=0; f<8; f++) {
			init_pair(b*8+f, colors[f], colors[b]);
		}
	}
	cbreak();
	noecho();
	nodelay(stdscr, TRUE);
	keypad(scr, TRUE);

	atexit((void (*)(void)) curses_end_video);
}

void curses_set_blink(int b) {
	b = b;	/* ignore */
}

void curses_flush_video(void)
{
	refresh();
}

int printable(int x) {
	if (x == '\n' || x == '\r' || x == '\b') return x;
	if (x < 0x20 || x > 126) return '?';
	return x;
}

void curses_put_cursor(unsigned x, unsigned y)
{
	int	c;

	if (x == 0) {
		c = mvinch(y, 0);
		mvaddch(y, 0, c);
		addch('\b');
	}
	else {
		c = mvinch(y, x-1);
		mvaddch(y, x-1, c);
	}
}

void curses_unput_cursor(void)
{
}

void curses_copy(unsigned x1, unsigned y1, unsigned x2, unsigned y2,
		  unsigned nx, unsigned ny)
{
	/* Cant figure this out.  Be evil. */
	int x, y;

	for (y=0; y<=y2-y1; y++) {
		for (x=0; x<=x2-x1; x++) {
			int at = mvinch(y1+y, x1+x);
			mvaddch(ny+y, nx+x, at);
		}
	}
}

void curses_new_cursor(int st, int end)
{
}

void curses_window_size(unsigned width, unsigned height)
{
}

void curses_draw_line(unsigned x, unsigned y, const char *text, unsigned len, BYTE attr)
{
	char	buf[101];
	int	i;

	if (len > 100) {
		fprintf(stderr, "Error: text too long in curses_draw_line()\n");
		exit(1);
	}
	for (i=0; i<len; i++)
		buf[i] = printable(text[i]);
	buf[i] = 0;
	attr = ((attr & 0x70) >> 1) | (attr & 0x07);
	color_set(attr, NULL);
	mvaddnstr(y, x, buf, len);
}

static BYTE scan_table1[256 - 0x20] =
{
	0x39, 0x02, 
#ifdef KBUK             /* double quotes, hash symbol */
	0x03, 0x2b,
#else
	0x28, 0x04,
#endif
	0x05, 0x06, 0x08, 0x28,
	0x0a, 0x0b, 0x09, 0x0d, 0x33, 0x0c, 0x34, 0x35,
	0x0b, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
	0x09, 0x0a, 0x27, 0x27, 0x33, 0x0d, 0x34, 0x35,
#ifdef KBUK             /* at symbol */
	0x28, 
#else
	0x03,
#endif
	0x1e, 0x30, 0x2e, 0x20, 0x12, 0x21, 0x22,
	0x23, 0x17, 0x24, 0x25, 0x26, 0x32, 0x31, 0x18,
	0x19, 0x10, 0x13, 0x1f, 0x14, 0x16, 0x2f, 0x11,
	0x2d, 0x15, 0x2c, 0x1a, 
#ifdef KBUK             /* backslash */
	0x56, 
#else
	0x2b,
#endif
	0x1b, 0x07, 0x0c,
	0x29, 0x1e, 0x30, 0x2e, 0x20, 0x12, 0x21, 0x22,
	0x23, 0x17, 0x24, 0x25, 0x26, 0x32, 0x31, 0x18,
	0x19, 0x10, 0x13, 0x1f, 0x14, 0x16, 0x2f, 0x11,
	0x2d, 0x15, 0x2c, 0x1a, 
#ifdef KBUK             /* vertical bar */
	0x56, 
#else
	0x2b,
#endif

	0x1b, 

#ifdef KBUK             /* tilde */
	0x2b,
#else
	0x29,
#endif
	0
};

static struct {
	int code;
	int scan;
} basic_keys[] = {
	{ 9,                 0x0f   },
	{ 10,                0x1c   },
	{ 27,                0x01   },
	{ KEY_BACKSPACE,     0x0e   },
	{ KEY_HOME,          0x47e0 },
	{ KEY_LEFT,          0x4be0 },
	{ KEY_UP,            0x48e0 },
	{ KEY_RIGHT,         0x4de0 },
	{ KEY_DOWN,          0x50e0 },
	{ KEY_PPAGE,         0x49e0 },
	{ KEY_NPAGE,         0x51e0 },
	{ KEY_SELECT,        0x4fe0 },
	{ KEY_IC,            0x52e0 },
	{ KEY_DC,            0x53e0 },
	{ KEY_F(1),          0x3be0 },
	{ KEY_F(2),          0x3ce0 },
	{ KEY_F(3),          0x3de0 },
	{ KEY_F(4),          0x3ee0 },
	{ KEY_F(5),          0x3fe0 },
	{ KEY_F(6),          0x40e0 },
	{ KEY_F(7),          0x41e0 },
	{ KEY_F(8),          0x42e0 },
	{ KEY_F(9),          0x43e0 },
	{ KEY_F(10),         0x44e0 }
};

int xlate_basic_key(int key) {
	int	i;

	for (i=0; i<(sizeof(basic_keys)/sizeof(basic_keys[0])); i++) {
		if (basic_keys[i].code == key)
			return basic_keys[i].scan;
	}
	return 0;
}

static unsigned translate(int key)
{
	int c;
	if (0x20 <= key && key <= 0x20+sizeof(scan_table1))
		return (scan_table1[key - 0x20]);
	c = xlate_basic_key(key);
	if (c != 0) return c;
	return key;
}

int shifted(int key) {
	char	s[] = "!@#$%^&*()_+|~QWERTYUIOP{}ASDFGHJKL:\"ZXCVBNM<>?";
	return strchr(s, key)? 1: 0;
}

int controld(int key) {
	return xlate_basic_key(key) == 0 && key < 0x20;
}

void curses_process_kbd_events(void)
{
	int got = getch();
	int scan, count;
	char buffer[10];

	if (got != ERR) {
		if ((scan = translate(got)) != 0) {
			/*
			 * Should really release SHIFT and CTRL
			 * only if they have been pressed before.
			 */
			if (shifted(got))
				scan = scan << 8 | 0x2A;
			else
				scan = scan << 8 | 0xAA;
			if (controld(got)) {
				scan = translate(got+0x40);
				scan = scan << 8 | 0x1D;
			}
			else {
				scan = scan << 8 | 0x9D;
			}
			for (count = 0; scan; count++) {
				buffer[count] = scan & 0xff;
				scan >>= 8;
			}

			if (port60_buffer_ok(count)) {
				D(printf("Returning %d scan code bytes\n",count););
				put_scancode(buffer, count);
			}
		}
	}
}
